/*
  RFIDuino Lock

  Sketch for an Arduino Uno to open an electronic door lock using a serial
  RFID reader.

  The RFID reader must output tag UIDs in hexadecimal, followed by CR or LF.

  The door opens when the right Arduino pin is set high for a brief moment.

  The UIDs are saved in EEPROM. A maximum of 15 UIDs with a maximum length of
  31 hexadecimal characters each may be saved.

  The current mode is indicated by 2 LEDs: the first one is on pin 13, where
  most Arduinos have a LED connected on the main board. The second one mirrors
  the first one on another pin, to drive another LED mounted on an enclosure for
  example.

  Use:

  - Power up the Arduino: the LEDs flash once. The Arduino reads UIDs from the
    serial line and opens the door when it matches one of the enrolled UIDs.

  - To enroll a UID, press RESET once without powering down the Arduino: after
    rebooting, the LEDs flash twice, indicating it's in UID enrollment mode.
    Present the tags to enroll to the RFID reader.

  - To delete enrolled UIDs, press RESET again without powering down the
    Arduino: after rebooting, the LEDs flash thrice, indicating it's in UID
    deletion mode. Present the tags to delete to the RFID reader.

  - To return to normal mode, press RESET again without powering down the
    Arduino, or power it down and then back up.
*/

#include <EEPROM.h>

//// Defines
#define LED1 13			// Pin 13 has a LED connected on most Arduinos
#define LED2 6			// Pin 6 mirrors LED1

#define DOOR_UNLOCK 2		// Pin 2 unlocks the door
#define DOOR_UNLOCK_DELAY 250	// How many ms the door unlock pin should be up

#define UID_MAX_LEN 32		// UIDs are max 31 printable char strings,
				// null-terminated
#define MAX_NB_UIDS 16		// Maximum 16 UIDs in EEPROM (16*32 = 512 chars)
#define EEFMT_MAGIC "FORMATTED"	// UID #0 should contains this magic to indicate
				// the EEPROM is formatted
#define RESET_MAGIC 0xdeadbeef  // To distinguish a power-on from a reset event

#define LOCK_DRIVE_MODE 0	// Normal mode: read UIDs and unlock the door
#define UID_ENROLL_MODE 1	// UID enroll mode
#define UID_DELETE_MODE 2	// UID delete mode

#define MODE_LED_FLASH_DUR 500	// How many ms the LEDs should flash ON and OFF
				// when indicating the mode
#define LED_ON_DUR 50		// Duration of LED flashes in normal mode
#define LED_OFF_DUR 1950	// Duration of pauses between LED flashes
				// in normal mode

#define UNLOCK_WAIT_QUIET 2000	// How many ms the serial port should stay
				// quiet before we accept a new UID after a
				// door unlock
//#define DEBUG



//// Reset-safe variables
uint32_t reset_magic __attribute__ ((section (".noinit")));
byte mode __attribute__ ((section (".noinit")));



//// Global variables
char uids[MAX_NB_UIDS][UID_MAX_LEN];

byte led_state = 0;

char uid[UID_MAX_LEN];
int uid_len;
byte uid_valid;



//// Routines

// EEPROM UID slot read routine
void read_eeprom_uid(int uidno) {
  int i, j;
  char c;

  j = 0;
  for(i = uidno * UID_MAX_LEN; i < (uidno + 1) * UID_MAX_LEN - 1; i++) {
    c = EEPROM.read(i);
    if(c >= 32 && c < 127)
      uids[uidno][j++] = c;
    else
      if(c == 0)
        break;
  }
  uids[uidno][j] = 0;
}

// EEPROM UID slot write routine
void write_eeprom_uid(int uidno) {
  int i, j;
  char c;

  j = uidno * UID_MAX_LEN;
  for(i = 0; i < UID_MAX_LEN - 1; i++) {
    c = uids[uidno][i];
    if(c >= 32 && c < 127) {
      if(EEPROM.read(j) != c)
        EEPROM.write(j, c);
      j++;
    }
    else
      if(c == 0)
        break;
  }
  if(EEPROM.read(j))
    EEPROM.write(j, 0);
}

// Mode LED flashing sequence
void mode_led_flash(void)
{
  int i;

  for(i = 0; i < mode + 1; i++) {
    digitalWrite(LED1, HIGH);
    digitalWrite(LED2, HIGH);
    delay(MODE_LED_FLASH_DUR);
    digitalWrite(LED1, LOW);
    digitalWrite(LED2, LOW);
    delay(MODE_LED_FLASH_DUR);
  }
}

// Setup routine
void setup() {
  int i;

  // Initialize the LEDs and the door unlock pins as outputs
  pinMode(LED1, OUTPUT);
  pinMode(LED2, OUTPUT);
  pinMode(DOOR_UNLOCK, OUTPUT);

  // Initialize the serial port
  Serial.begin(9600);

  // Read the first UID in EEPROM: if the magic isn't there, format the EEPROM
  read_eeprom_uid(0);
  if(strcmp(uids[0], EEFMT_MAGIC)) {
#ifdef DEBUG
    Serial.println("EEPROM not formatted");
#endif
    strcpy(uids[0], EEFMT_MAGIC);
    for(i = 1; i < MAX_NB_UIDS; i++)
      uids[i][0] = 0;
    for(i = 0; i < MAX_NB_UIDS; i++) {
#ifdef DEBUG
      Serial.println("Formatting EEPROM UID slot " + String(i));
#endif
      write_eeprom_uid(i);
    }
  }

  // Read the rest of the EEPROM if it's formatted
  else
  {
#ifdef DEBUG
    Serial.println("EEPROM formatted:");
#endif
    for(i = 1; i < MAX_NB_UIDS; i++) {
      read_eeprom_uid(i);
#ifdef DEBUG
    Serial.println("UID " + String(i) + ": " + uids[i]);
#endif
    }
  }

  // Determine the working mode:
  // - If the board was powered on, go into normal mode
  // - If the board was reset with the reset button, cycle between
  //   UID enroll mode, delete mode and normal mode
  if(reset_magic != RESET_MAGIC) {
#ifdef DEBUG
    Serial.println("LOCK drive mode");
#endif
    reset_magic = RESET_MAGIC;
    mode = LOCK_DRIVE_MODE;
  }
  else
    switch(mode) {
      case LOCK_DRIVE_MODE:
#ifdef DEBUG
        Serial.println("UID enroll mode");
#endif
        mode = UID_ENROLL_MODE;
        break;
      case UID_ENROLL_MODE:
#ifdef DEBUG
        Serial.println("UID delete mode");
#endif
        mode = UID_DELETE_MODE;
        break;
      default:
#ifdef DEBUG
        Serial.println("LOCK drive mode");
#endif
        mode = LOCK_DRIVE_MODE;
        break;
    }

  // Flash the mode on the LED
  mode_led_flash();

  // Initialize the serial receive buffer
  uid[0] = 0;
  uid_len = 0;
  uid_valid = 1;
}

// Loop routine
void loop() {
  char c;
  int match_uid_i;

  // Turn the LEDs ON or OFF and set the timeout for the duration of a flash or
  // a pause if no serial characters are received
  digitalWrite(LED1, led_state);
  digitalWrite(LED2, led_state);
  Serial.setTimeout(led_state ? LED_ON_DUR : LED_OFF_DUR);

  // Receive serial characters
  if(Serial.readBytes(&c, 1)) {

    // Check that the character is a hex digit and add it to the buffer
    // if it is and there's space in the buffer, else invalidate the buffer
    if(((c >= 0x30 && c < 0x3a) ||
	(c >= 0x41 && c < 0x47) ||
	(c >= 0x61 && c < 0x67))) {
      if(uid_valid && uid_len < UID_MAX_LEN - 1) {
        uid[uid_len++] = toupper(c);
        uid[uid_len] = 0;
      }
      else {
        uid_valid = 0;
      }
    }
    else {

      // Check that the character is CR or LF
      if(c == 0x0a || c == 0x0d) {

        // If the buffer is valid and not empty, we have a UID
        if(uid_valid && uid_len) {

#ifdef DEBUG
          Serial.println("Received UID: " + String(uid));
#endif
          // Find a matching UID in the list of enrolled UID
          for(match_uid_i = 0;
		match_uid_i < MAX_NB_UIDS && strcmp(uids[match_uid_i], uid);
		match_uid_i++);

          // Do we have a match?
          if(match_uid_i < MAX_NB_UIDS) {

#ifdef DEBUG
            Serial.println("UID match at position " + String(match_uid_i));
#endif
            // What mode are we in?
            switch(mode) {

              case LOCK_DRIVE_MODE:	// Normal mode: open the door

#ifdef DEBUG
                Serial.println("Door unlock");
#endif
                // Turn on the LEDs
                digitalWrite(LED1, HIGH);
                digitalWrite(LED2, HIGH);

                // Drive the door unlock pin high for a short period of time
                digitalWrite(DOOR_UNLOCK, HIGH);
                delay(DOOR_UNLOCK_DELAY);
                digitalWrite(DOOR_UNLOCK, LOW);

                // Wait until the serial port stays quiet long enough, to avoid
		// multiple unlocks with repeating RFID readers
                Serial.setTimeout(UNLOCK_WAIT_QUIET);
                while(Serial.readBytes(&c, 1));

                // Turn off the LEDs
                digitalWrite(LED1, LOW);
                digitalWrite(LED2, LOW);

                break;

              case UID_DELETE_MODE:	// UID delete mode: delete the UID

#ifdef DEBUG
                Serial.println("UID deleted");
#endif
                // Erase the UID from this slot
                uids[match_uid_i][0] = 0;
                write_eeprom_uid(match_uid_i);

                // Flash the mode on the LEDs to indicate successful deletion
                mode_led_flash();

                break;
            }
          }

          // We don't have a match
          else {

#ifdef DEBUG
            Serial.println("No UID match");
#endif
            // Are we in UID enroll mode?
            if(mode == UID_ENROLL_MODE) {

              // Find the first empty slot in the list of UIDs
              for(match_uid_i = 0;
			match_uid_i < MAX_NB_UIDS && uids[match_uid_i][0];
			match_uid_i++);

              // Did we find a free slot?
              if(match_uid_i < MAX_NB_UIDS) {

#ifdef DEBUG
                Serial.println("Free UID slot found at position " +
				String(match_uid_i) + ": UID added");
#endif
                // Add the UID in this slot
                strcpy(uids[match_uid_i], uid);
                write_eeprom_uid(match_uid_i);

                // Flash the mode on the LEDs to indicate successful enrollment
                mode_led_flash();
              }
            }
          }
        }

        // Reset the buffer
        uid[0] = 0;
        uid_len = 0;
        uid_valid = 1;
      }

      else {
        // Any character other than a hex digit, CR, LF or STX / ETX
        // (which some readers send) invalidates the buffer
        if(c != 0x02 && c != 0x03)
          uid_valid = 0;
      }
    }
  }

  // Change the state of the LEDs
  led_state ^= 1;
}
